from .basic.constants import *
from .construct_model.model import Net
from .train_method import train_step, test_step
from .construct_model.dataset import OptimizeDataset, TestDataset, CustomDataset, SearchDataset
from .basic.loss_fn import LogLoss, LogL1Loss
from .test_method import visualise_model, model_input, path_input


def optimize(path, model, learning_rate=1e-4, epochs=1000000, check_interval=100):
    model.load_state_dict(torch.load(path))

    loss_fn = LogLoss()
    optimizer = torch.optim.Adam(model.parameters(), lr=learning_rate)

    # Create dataset instances
    train_dataset = SearchDataset()
    test_dataset = TestDataset()
    train_losses = []
    test_losses = []

    for t in range(epochs + 1):
        # Test
        test_loss = test_step(test_dataset, model, loss_fn)

        # Train
        train_loss = train_step(train_dataset, model, loss_fn, optimizer)
        if train_loss == -1:
            break

        train_losses.append(train_loss.item())
        test_losses.append(test_loss.item())
        if t % 500 == 0 and t != 0:
            # print(f"\n-----Epoch {t}------")
            # print(f"Train Error: \n Avg loss: {np.mean(train_losses[-500:]):>8f}")
            # print(f"Test Error: \n Avg loss: {np.mean(test_losses[-500:]):>8f}")
            torch.save(model.state_dict(), 'best_model.pth')
            # print("*** model saved! ***")

        if t % check_interval == 0 and t != 0:
            # fig, ax = plt.subplots(1, 2, figsize=(12, 6))
            # ax[0].plot([loss for loss in test_losses if loss < 500], label='test_loss')
            # ax[1].plot([loss for loss in train_losses if loss < 500], label='train_loss')
            # plt.savefig(f'./optimizing/{t / 100000}.jpg')

            print(visualise_model(path_input('./best_model.pth'), print_params=False, draw=True))
            # name1 = visualise_model(model_input('./best_models/k/4.88_4.23_2.44.pth'), print_params=False, draw=True)
            # plt.legend()
            # plt.show()



from .constants import *


def J_sm(w):
    """
    :param w: numpy array
    :return: single tensor value
    """

    J_positive = (2 * g_const ** 2 / np.pi *
                  (kappa * w_c * w /
                   ((w_c ** 2 - w ** 2) ** 2 + kappa ** 2 * w ** 2)))
    # 阶跃函数
    J = np.where(w >= w_0, J_positive, 0)
    return torch.tensor(J, dtype=torch.float64)


# 定义目标函数 J_tar
def J_tar(w):
    """
    :param w: numpy array
    :return: single tensor value
    """
    J = np.float64(
        np.where(
            w >= w_0,
            J_sm(w),
            J_t + k / (w_cc - w) ** 20
        ))
    return torch.tensor(J, dtype=torch.float64)


import torch


def J_mod(w, g_vec, w_vec, k_vec):
    """
    :param w: numpy array
    :param g_vec: tensor array
    :param w_vec: tensor array
    :param k_vec: tensor array
    :return: tensor array as w
    """
    w = torch.tensor(w, dtype=torch.float64)

    # omega_ij:H矩阵的实部
    tril_indices = torch.tril_indices(n_mod, n_mod, offset=0)
    w_lower = torch.zeros((n_mod, n_mod), dtype=torch.float64)
    w_lower[tril_indices[0], tril_indices[1]] = w_vec

    real_part = w_lower.t() + w_lower  # 对称化

    # kappa_ij：H矩阵的虚部
    imag_part = -torch.diag(k_vec) / 2

    # H矩阵
    H_mat = torch.complex(real_part, imag_part)

    # 添加batch维度
    H_mat = H_mat.unsqueeze(0).expand(w.shape[0], -1, -1)

    # omega * I
    eye = torch.eye(n_mod, dtype=torch.float64).unsqueeze(0).expand_as(H_mat)
    w_eye = w.view(-1, 1, 1) * eye

    # H - omega * I
    H_w_mat = H_mat - w_eye

    # 计算逆矩阵
    inverse_mat = torch.inverse(H_w_mat)

    # 取虚部
    imag_mat = inverse_mat.imag

    # g' * Im(H-w) * g
    g_vec_mat = g_vec.view(1, -1, 1).expand(w.shape[0], -1, 1)
    result = torch.bmm(torch.bmm(g_vec_mat.transpose(1, 2), imag_mat), g_vec_mat)

    return (result / pi).view(-1)


def inverse_first_column(matrix):
    """
    只计算逆矩阵的第一列

    :param matrix: The input matrix (torch tensor) with batch dimension.
    :return: The first column of the inverse matrix for each batch.
    """
    identity = torch.eye(n_mod, dtype=torch.float64).unsqueeze(0).expand(matrix.shape[0], -1, -1)
    first_column = torch.linalg.solve(matrix, identity[:, :, 0])
    return first_column.unsqueeze(2)


def inverse_first_row(matrix):
    """
    只计算逆矩阵的第一行

    :param matrix: The input matrix (torch tensor) with batch dimension.
    :return: The first row of the inverse matrix for each batch as column vectors.
    """
    identity = torch.eye(n_mod, dtype=torch.float64).unsqueeze(0).expand(matrix.shape[0], -1, -1)
    first_row = torch.linalg.solve(matrix, identity[:, 0, :])
    return first_row.unsqueeze(1)


def single_k_J_mod(w, g_vec, w_vec, k):
    """
    :param w: numpy array
    :param g_vec: tensor array
    :param w_vec: tensor array
    :param k: tensor array
    :return: tensor array as w
    """
    # 计算 w_ij - omega (H - omega 实部)
    tril_indices = torch.tril_indices(n_mod, n_mod, offset=0)
    w_ij = torch.zeros((n_mod, n_mod), dtype=torch.float64)
    w_ij[tril_indices[0], tril_indices[1]] = w_vec

    w_ij_mat = w_ij.t() + w_ij  # 对称化

    w_ij_mat = w_ij_mat.unsqueeze(0).expand(w.shape[0], -1, -1)  # 添加 batch 维度

    eye = torch.eye(n_mod, dtype=torch.float64).unsqueeze(0).expand_as(w_ij_mat)
    w_eye = torch.tensor(w, dtype=torch.float64).view(-1, 1, 1) * eye
    w_w_mat = w_ij_mat - w_eye

    # 计算逆矩阵虚部
    kk = k / 2
    inverse_column = inverse_first_column(w_w_mat)
    temp_mat = w_w_mat.clone()
    temp_mat[:, 0, 0] = temp_mat[:, 0, 0] + inverse_column[:, 0, 0] * (kk ** 2)
    inverse_row = inverse_first_row(temp_mat)
    imag_mat = torch.bmm(inverse_column, inverse_row) * kk

    # print(f'single k :{imag_mat[0]}')
    print(temp_mat[0])

    # g' * Im(H-w) * g
    g_vec_mat = g_vec.view(1, -1, 1).expand(w.shape[0], -1, 1)
    result = torch.bmm(torch.bmm(g_vec_mat.transpose(1, 2), imag_mat), g_vec_mat)

    return (result / pi).view(-1)

# # 计算逆矩阵虚部
#     kk = k / 2
#     inverse_column = inverse_first_column(H_w_mat)
#     temp_mat = H_w_mat.clone()
#     temp_mat[:, 0, 0] = temp_mat[:, 0, 0] + inverse_column[:, 0, 0] * kk ** 2
#     inverse_row = inverse_first_row(temp_mat)
#
#     # g' * Im(H-w) * g
#     g_vec_mat = g_vec.view(1, -1, 1).expand(w.shape[0], -1, 1)
#     result = torch.bmm(g_vec_mat.transpose(1, 2), inverse_column) * torch.bmm(inverse_row, g_vec_mat) * kk



from torch import nn

from ..basic.constants import *
from ..basic.J_fn import J_mod, single_k_J_mod

g_vec = torch.ones(n_mod, dtype=torch.float64) * 0.1
w_vec = torch.rand(int(n_mod * (n_mod + 1) / 2), dtype=torch.float64) * 8 - 4
w_vec[diagonal_idx] = torch.rand(len(diagonal_idx), dtype=torch.float64) * 5
k_vec = torch.ones(n_mod, dtype=torch.float64) * 5


class Net(nn.Module):
    def __init__(self, g_vec=g_vec, w_vec=w_vec, k_vec=k_vec):
        super().__init__()

        # 初始化参数
        self.g_vec = nn.Parameter(g_vec)
        self.w_vec = nn.Parameter(w_vec)
        self.k_vec = nn.Parameter(k_vec)

    def forward(self, w):
        return J_mod(w, self.g_vec, self.w_vec, self.k_vec)


class SingleKNet(nn.Module):
    def __init__(self, g_vec=g_vec, w_vec=w_vec, k_vec=k_vec):
        super().__init__()

        # 初始化参数
        self.g_vec = nn.Parameter(g_vec)
        self.w_vec = nn.Parameter(w_vec)
        self.k_vec = nn.Parameter(k_vec)

    def forward(self, w):
        return single_k_J_mod(w, self.g_vec, self.w_vec, self.k)



from torch.optim.lr_scheduler import ReduceLROnPlateau

from .basic.constants import *
from .train_method import train_step, test_step
from .construct_model.dataset import SearchDataset, TestDataset
from .basic.loss_fn import LogLoss
from .construct_model.model import Net, SingleKNet
from .test_method import visualise_model, model_input, path_input


def crude_train(model, train_dataset, test_dataset, loss_fn, optimizer, scheduler, model_number):
    """
    这个函数用于粗糙地对随机模型进行梯度下降
    返回最终的 loss 以反映粗糙训练后模型的质量
    :param model: 未训练的模型
    :return: 训练完成后的 loss
    """
    epoch_counter = 0
    while optimizer.param_groups[0]['lr'] > min_lr:
        # Test
        test_loss = test_step(test_dataset, model, loss_fn)

        # Train
        train_loss = train_step(train_dataset, model, loss_fn, optimizer)
        if train_loss == -1:
            print("\n!!! loss = NaN ,model dumped !!!\n")
            break

        # print every 500 epoch
        if epoch_counter % 1000 == 0:
            print(f"\n---Model {model_number} Epoch {epoch_counter}---")
            print(f"Train Error: \n Avg loss: {train_loss:>8f}")
            print(f"Test Error: \n Avg loss: {test_loss:>8f}")
            if epoch_counter == 15000:
                model.k_vec.requires_grad = True
            torch.save(model.state_dict(), './best_model.pth')

        # Update learning rate
        scheduler.step(test_loss)

        epoch_counter += 1

    print("\n！！！Stopping training！！！\n")
    return model


if __name__ == '__main__':
    learning_rate = 1e-3  # 初始学习率
    min_lr = 1e-5  # 最小学习率

    train_dataset = SearchDataset()
    test_dataset = TestDataset()

    # 每次循环训练一个随机初值模型
    model_counter = 0
    while True:
        print(f'### Begin to train model No.{model_counter+1} ###')
        # 初始化
        model = Net()

        loss_fn = LogLoss()
        optimizer = torch.optim.Adam([
            {'params': model.g_vec, 'lr': learning_rate / 10},
            {'params': model.w_vec, 'lr': learning_rate},
            {'params': model.k_vec, 'lr': learning_rate / 10}
        ])
        scheduler = ReduceLROnPlateau(
            optimizer,
            mode='min',
            factor=0.1,
            patience=1000,
            min_lr=min_lr,
            verbose=True,
        )

        # 粗糙训练
        trained_model = crude_train(
            model,
            train_dataset,
            test_dataset,
            loss_fn,
            optimizer,
            scheduler,
            model_number=model_counter+1,
        )

        # 保存优秀模型
        torch.save(trained_model.state_dict(), './best_model.pth')
        name, loss_pos, loss_neg = visualise_model(path_input('./best_model.pth'), draw=True)
        torch.save(trained_model.state_dict(), f'./models/1.5/{name}.pth')

        model_counter += 1




import torch
from torch import nn


class LogLoss(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, w, pred, target):
        w = torch.tensor(w, dtype=torch.float64)
        log = torch.log(target / pred)

        # if torch.isnan(log).any():
        #     nan_indices = torch.nonzero(torch.isnan(torch.log(target / pred))).flatten()
        #     print("w = ", w[nan_indices].tolist())
        #     print("pred = ", pred[nan_indices].tolist())
        #     print("target = ", target[nan_indices].tolist())
        #     print("target / pred = ", (target / pred)[nan_indices].tolist())
        #     return -1

        non_nan_log = log[~torch.isnan(log)]
        if non_nan_log.numel() > 0:
            max_log = torch.max(non_nan_log)
            log[torch.isnan(log)] = max_log * 2
            loss1 = torch.sum(
                torch.where(
                    w >= 0,
                    log ** 2,
                    torch.zeros_like(pred)
                )
            )
            loss2 = torch.sum(
                torch.where(
                    w < 0,
                    torch.where(
                        pred > target,
                        torch.ones_like(pred),
                        torch.zeros_like(pred)
                    ) * log ** 2,
                    torch.zeros_like(pred)
                )
            )
            loss = torch.sqrt(loss1 + loss2)
            return loss
        else:
            return -1


class LogL1Loss(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, w, pred, target):
        w = torch.tensor(w, dtype=torch.float64)
        log = torch.log(target / pred)

        # if torch.isnan(log).any():
        #     nan_indices = torch.nonzero(torch.isnan(torch.log(target / pred))).flatten()
        #     print("w = ", w[nan_indices].tolist())
        #     print("pred = ", pred[nan_indices].tolist())
        #     print("target = ", target[nan_indices].tolist())
        #     print("target / pred = ", (target / pred)[nan_indices].tolist())
        #     return -1

        non_nan_log = log[~torch.isnan(log)]
        if non_nan_log.numel() > 0:
            max_log = torch.max(non_nan_log)
            log[torch.isnan(log)] = max_log * 2
            loss1 = torch.sum(
                torch.where(
                    w >= 0,
                    log ** 2,
                    torch.zeros_like(pred)
                )
            )
            loss2 = torch.sum(
                torch.where(
                    w < 0,
                    torch.where(
                        pred > target,
                        torch.ones_like(pred),
                        torch.zeros_like(pred)
                    ) * log ** 2,
                    torch.zeros_like(pred)
                )
            )
            loss3 = torch.sum(
                torch.where(
                    (pred - target) ** 2 > 1e-5,
                    torch.abs(pred - target),
                    torch.zeros_like(pred)
                )
            )
            loss = torch.sqrt(loss1 + loss2) + loss3
            return loss
        else:
            return -1


def PadeLog24(x):
    PadeLog = (-1 + x) ** 2 / (20 * (-1 + x) ** 2 - (-1 + x) ** 4 + 240 * x)
    Linear = 1.58325 - 428.203 * x
    return torch.where(x >= -0.09, PadeLog, Linear)


def PadeLog22(x):
    PadeLog = (x - 1) ** 2 / ((x - 1) ** 2 / 12 + x)
    Linear = 12 - 144 * x
    return torch.where(x >= 0, PadeLog, Linear)


class PadeLoss(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, w, pred, target):
        # 获取pred中的零元素索引
        zero_indices = torch.nonzero(pred == 0).flatten()

        # 将pred中的零元素替换为1e-6
        pred[zero_indices] = 1e-6

        PadeLog = PadeLog22(target / pred)

        loss1 = torch.sum(
            torch.where(
                w >= 0,
                PadeLog,
                torch.zeros_like(pred)
            )
        )
        loss2 = torch.sum(
            torch.where(
                w < 0,
                torch.where(
                    pred > target,
                    torch.ones_like(pred),
                    torch.zeros_like(pred)
                ) * PadeLog,
                torch.zeros_like(pred)
            )
        )
        loss = torch.sqrt(loss1 + loss2)
        if torch.isnan(loss):
            print("*** Error ***\n    loss = NaN!")
            nan_indices = torch.nonzero(torch.isnan(PadeLog22(target / pred))).flatten()
            print("    w = ", w[nan_indices].tolist())
            print("    pred = ", pred[nan_indices].tolist())
            print("    target / pred = ", (target / pred)[nan_indices].tolist())
            print("    PadeLog(target / pred) = ", (PadeLog22(target / pred))[nan_indices].tolist())
            exit()
        else:
            return loss



from .basic.constants import *


def apply_bound(model):
    with torch.no_grad():
        for name, param in model.named_parameters():
            if name == 'g_vec':
                param.data.clamp_(min=0)
            elif name == 'w_vec':
                idx = 0
                for w_ij in param.data.view(-1):
                    if idx in diagonal_idx:  # 对角线元素
                        w_ij.clamp_(min=0)
                    else:
                        w_ij.clamp_(min=-30, max=30)
                    idx += 1
            elif name == 'k_vec':
                param.data.clamp_(min=0)


def train_step(dataset, model, loss_fn, optimizer):
    model.train()
    w, J = dataset[0]

    # 前向传播
    pred = model(w)

    # 计算损失
    loss = loss_fn(w, pred, J)

    # 反向传播
    loss.backward()

    # 更新参数
    optimizer.step()
    optimizer.zero_grad()

    # 应用参数限制
    apply_bound(model)

    return loss


def test_step(dataset, model, loss_fn):
    model.eval()
    w, J = dataset[0]
    pred = model(w)
    test_loss = loss_fn(w, pred, J)
    return test_loss




from torch import nn

from ..basic.J_fn import J_tar, J_mod
from ..basic.loss_fn import LogLoss
from model import Net
from ..basic.constants import *

evaluate_input = np.concatenate((
    np.linspace(-3, -0.08, 10 * 5),
    np.linspace(-0.04, 0.4, 25 * 10),
    np.linspace(0.405, 1, 20 * 10),
    np.linspace(1.05, 3, 25 * 5)
))


def cat_input(cat_param):
    """
    这个函数用于将合并在一起的 params 差分
    :param cat_param: 合并的 params
    :return: 拆分的 params ([torch]格式)
    """
    len_g_vec = n_mod
    len_w_vec = int(n_mod * (n_mod + 1) / 2)
    len_k_vec = n_mod

    g_vec = cat_param[:len_g_vec]
    w_vec = cat_param[len_g_vec:len_g_vec + len_w_vec]
    k_vec = cat_param[len_g_vec + len_w_vec:len_g_vec + len_w_vec + len_k_vec]

    g_vec = torch.tensor(g_vec)
    w_vec = torch.tensor(w_vec)
    k_vec = torch.tensor(k_vec)

    return [g_vec, w_vec, k_vec]


def path_input(path):
    """
    这个函数用于提取模型中的参数
    :param path: 模型地址
    :return: 拆分的模型参数（[torch]格式）
    """
    model = Net()
    model.load_state_dict(torch.load(path))

    params_dict = {}
    for name, param in model.named_parameters():
        params_dict[name] = param.clone().detach()

    g_vec = params_dict['g_vec']
    w_vec = params_dict['w_vec']
    k_vec = params_dict['k_vec']

    return [g_vec, w_vec, k_vec]


def model_input(model):
    """
    这个函数用于提取模型中的参数
    :return: 拆分的模型参数（[torch]格式）
    """
    params_dict = {}
    for name, param in model.named_parameters():
        params_dict[name] = param.clone().detach()

    g_vec = params_dict['g_vec']
    w_vec = params_dict['w_vec']
    k_vec = params_dict['k_vec']

    return [g_vec, w_vec, k_vec]


def visualise_model(param_list, evaluate_w=evaluate_input, print_params: bool = False, draw: bool = True):
    """
    这个函数用于
    画给定模型的拟合曲线
     print 参数
    计算模型 loss
    """
    g_vec = param_list[0]
    w_vec = param_list[1]
    k_vec = param_list[2]

    # 计算 loss
    target = J_tar(evaluate_w)
    pred = J_mod(evaluate_w, g_vec, w_vec, k_vec)
    loss_fn = LogLoss()
    loss = loss_fn(evaluate_w, pred, target).item()

    pos_idx = np.where(evaluate_w > 0)[0]
    loss_pos = loss_fn(evaluate_w[pos_idx], pred[pos_idx], target[pos_idx]).item()

    mid_idx = np.where((0 < evaluate_w) & (evaluate_w < 0.4))[0]
    loss_mid = loss_fn(evaluate_w[mid_idx], pred[mid_idx], target[mid_idx]).item()

    neg_idx = np.where(evaluate_w < 0)[0]
    loss_neg = loss_fn(evaluate_w[neg_idx], pred[neg_idx], target[neg_idx]).item()

    # print(f'total loss:{round(loss, 2)}')
    # print(f'positive loss:{round(loss_pos, 2)}')
    # print(f'negative loss:{round(loss_neg, 2)}')
    # print(f'middle loss:{round(loss_mid, 2)}')
    name = f'{round(loss, 2)}_{round(loss_pos, 2)}_{round(loss_neg, 2)}'
    # print(name)

    # print 参数
    if print_params:
        print(f'g:{g_vec.tolist()}')
        print(f'w:{w_vec.tolist()}')
        print(f'k:{k_vec.tolist()}')

    # 画图
    if draw:
        w_plot = np.linspace(-3, 3, 1000)
        plt.plot(w_plot, J_tar(w_plot), label='target', linewidth=1)
        plt.plot(w_plot, J_mod(w_plot, g_vec, w_vec, k_vec), label='pred', linewidth=1)

        plt.yscale('log')
        plt.ylim([1e-10, 1])

    return name, loss_pos, loss_neg


def reorder(param_list, save_model: bool = False, path=''):
    """
    重新排列参数顺序
    """
    g_vec = param_list[0]
    w_vec = param_list[1]
    k_vec = param_list[2]

    # 检索 k_vec 中的非零值的索引
    idx = torch.where(k_vec != 0)
    if len(idx) != 1:
        raise ValueError("k_vec 中非零值的数量不为 1")
    else:
        # 交换非零元素与第一个元素的位置
        idx = idx[0].item()

        temp = g_vec[0].clone()
        g_vec[0] = g_vec[idx]
        g_vec[idx] = temp

        tril_indices = torch.tril_indices(n_mod, n_mod, offset=0)
        w_lower = torch.zeros((n_mod, n_mod), dtype=torch.float64)
        w_lower[tril_indices[0], tril_indices[1]] = w_vec
        w_ij = w_lower.t() + w_lower - torch.diag(torch.diag(w_lower))

        temp = w_ij[0].clone()
        w_ij[0] = w_ij[idx]
        w_ij[idx] = temp

        temp = w_ij[:, 0].clone()
        w_ij[:, 0] = w_ij[:, idx]
        w_ij[:, idx] = temp

        w_vec = w_ij[tril_indices[0], tril_indices[1]]

        k_vec[0] = k_vec[idx]
        k_vec[idx] = 0

        if save_model:
            model = Net()

            model.g_vec = nn.Parameter(g_vec)
            model.w_vec = nn.Parameter(w_vec)
            model.k_vec = nn.Parameter(k_vec)
            # print(path)
            torch.save(model.state_dict(), path.replace('.pth', '_reordered.pth'))
        return [g_vec, w_vec, k_vec]


if __name__ == '__main__':
    visualise_model(path_input(path='./best_model.pth'), print_params=True)
    # visualise_model(path_input(path='./best_models/k/4.66_3.62_2.93_reordered.pth'), print_params=True)
    plt.show()



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="HITphysics_curve_fit_package",
    version="0.0",
    author="Shangjing Liu",
    author_email="2022110002@stu.hit.edu.cn",
    description="A self-used package for curve fitting",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="",
    packages=setuptools.find_packages(),
    install_requires=['torch', 'numpy', 'matplotlib', 'simpy'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



from ..basic.constants import *
from ..basic.J_fn import J_tar


class SearchDataset:
    def __init__(self):
        input_array = np.concatenate((
            np.linspace(-3, -1, 5 * 3),
            np.linspace(-1, -0.08, 5 * 3),
            np.linspace(-0.04, 0.4, 20 * 6),
            np.linspace(0.405, 1, 20 * 6),
            np.linspace(1.05, 3, 30 * 3)
        ))

        # input_array = np.concatenate((
        #         np.linspace(-3, -0.5, 5 * 3),
        #         np.linspace(-0.5, -0.01, 50 * 3),
        #         np.linspace(-0.01, 0.05, 300 * 6),
        #         np.linspace(0.05, 1, 20 * 6),
        #         np.linspace(1.05, 3, 30 * 3)
        #     ))

        self.inputs = input_array
        self.outputs = J_tar(self.inputs)

    def __len__(self):
        return len(self.inputs)

    def __getitem__(self, idx):
        return self.inputs, self.outputs


class OptimizeDataset:
    def __init__(self):
        input_array = np.concatenate((
            np.linspace(-3, -1, 7 * 5),
            np.linspace(-1, -0.08, 11 * 5),
            np.linspace(-0.04, 0.4, 26 * 10),
            np.linspace(0.405, 1, 21 * 10),
            np.linspace(1.05, 3, 26 * 5)
        ))

        self.inputs = input_array
        self.outputs = J_tar(self.inputs)

    def __len__(self):
        return len(self.inputs)

    def __getitem__(self, idx):
        return self.inputs, self.outputs


class TestDataset:
    def __init__(self):
        input_array = np.concatenate((
            np.linspace(-3, -0.08, 10),
            np.linspace(-0.04, 0.4, 25),
            np.linspace(0.405, 1, 20),
            np.linspace(1.05, 3, 25)
        ))

        self.inputs = input_array
        self.outputs = J_tar(self.inputs)

    def __len__(self):
        return len(self.inputs)

    def __getitem__(self, idx):
        return self.inputs, self.outputs


class CustomDataset:
    def __init__(self, input_array):
        self.inputs = input_array
        self.outputs = J_tar(self.inputs)

    def __len__(self):
        return len(self.inputs)

    def __getitem__(self, idx):
        return self.inputs, self.outputs
